import java.io.PipedOutputStream;
import java.io.PipedInputStream;

import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.DataLine;
import javax.sound.sampled.SourceDataLine;



class Audio_Processing_Thread extends Thread{
	// Note that this value will influence how often pipe operations occur (a larger buffer will require less pipe
	// operations) and also how often notifications can be sent between the main and audio processing threads. The
	// buffer should probably be sized to be about large enough to store one or two frame durations worth of samples.
	private static final int BUFFER_SIZE=4096;

	// Flags to be used for indicating status to other threads and setting status from other threads.
	public volatile boolean is_Ready_To_Proceed=false;
	public volatile boolean should_Play_Audio=false;
	public volatile boolean should_Exit=false;

	public SourceDataLine audio_Line_To_Play_Audio_To;
	public AudioInputStream audio_Input_Stream_To_Copy_Downsampled_Audio_To;

	private AudioInputStream audio_Input_Stream;
	private AudioFormat decoded_Audio_Input_Stream_Format;
	private AudioInputStream decoded_Audio_Input_Stream;
	private PipedOutputStream decoded_Audio_Piped_Output_Stream;



	Audio_Processing_Thread(AudioInputStream audio_Input_Stream_To_Process){
		audio_Input_Stream=audio_Input_Stream_To_Process;


		// Get details about the audio_Input_Stream. For unspecified values, assign them a high quality setting, the
		// system probably should be able to convert the audio to that format.
		AudioFormat audio_Input_Stream_Format=audio_Input_Stream.getFormat();

		// TODO: If values are unspecified, try using the highest quality setting supported by the system.
		float audio_Input_Stream_Sample_Rate=audio_Input_Stream_Format.getSampleRate();
		if(audio_Input_Stream_Sample_Rate==AudioSystem.NOT_SPECIFIED) audio_Input_Stream_Sample_Rate=44100;

		int audio_Input_Stream_Sample_Size_In_Bits=audio_Input_Stream_Format.getSampleSizeInBits();
		if(audio_Input_Stream_Sample_Size_In_Bits==AudioSystem.NOT_SPECIFIED) audio_Input_Stream_Sample_Size_In_Bits=16;

		int audio_Input_Stream_Number_Of_Channels=audio_Input_Stream_Format.getChannels();
		if(audio_Input_Stream_Number_Of_Channels==AudioSystem.NOT_SPECIFIED) audio_Input_Stream_Number_Of_Channels=2;


		// Set up the decoded_Audio_Input_Stream.
		decoded_Audio_Input_Stream_Format=new AudioFormat(audio_Input_Stream_Sample_Rate,
		  audio_Input_Stream_Sample_Size_In_Bits, audio_Input_Stream_Number_Of_Channels, true, true);
		decoded_Audio_Input_Stream=AudioSystem.getAudioInputStream(decoded_Audio_Input_Stream_Format,
		  audio_Input_Stream);


		// Set up the audio_Line_To_Play_Audio_To.
		try{
			audio_Line_To_Play_Audio_To=(SourceDataLine)AudioSystem.getLine(new DataLine.Info(SourceDataLine.class,
			  decoded_Audio_Input_Stream_Format));
			audio_Line_To_Play_Audio_To.open(decoded_Audio_Input_Stream_Format);
		}catch(Exception e){
			System.out.println("An exception occurred while setting up audio line.");
		}


		// Set up the audio_Input_Stream_To_Copy_Downsampled_Audio_To.
		decoded_Audio_Piped_Output_Stream=new PipedOutputStream();
		try{
			audio_Input_Stream_To_Copy_Downsampled_Audio_To=AudioSystem.getAudioInputStream(
			  new AudioFormat(audio_Input_Stream_Sample_Rate, 8, 1, false, true),
			  new AudioInputStream(new PipedInputStream(decoded_Audio_Piped_Output_Stream, BUFFER_SIZE),
			  decoded_Audio_Input_Stream_Format, decoded_Audio_Input_Stream.getFrameLength()));
		}catch(Exception e){
			System.out.println("An exception occurred while setting up the downsampled audio stream.");
		}
	}



	public void run(){

		// Probably should make buffer at least large enough to store samples for one frame duration.
		byte[] buffer=new byte[BUFFER_SIZE];
		int number_Of_bytes_To_Copy, number_Of_Bytes_Read=0, number_Of_Bytes_Written=0;


		// Do an initial (but not necessarily full) fill of the audio buffer and pipe as necessary.
		try{
			// Determine the maximum available buffer space we have to use and use that value for the
			// number_Of_bytes_To_Copy.
			if(audio_Line_To_Play_Audio_To!=null){
				number_Of_bytes_To_Copy=audio_Line_To_Play_Audio_To.available();
				if(number_Of_bytes_To_Copy>BUFFER_SIZE) number_Of_bytes_To_Copy=BUFFER_SIZE;
			}else
				number_Of_bytes_To_Copy=BUFFER_SIZE;


			number_Of_bytes_To_Copy=number_Of_Bytes_Read=decoded_Audio_Input_Stream.read(buffer, 0,
			  number_Of_bytes_To_Copy);
			if(number_Of_Bytes_Read>0){
				if(audio_Line_To_Play_Audio_To!=null)
					number_Of_bytes_To_Copy=number_Of_Bytes_Written=audio_Line_To_Play_Audio_To.write(buffer, 0,
					  number_Of_bytes_To_Copy);

				// Note that at this point the main thread won't be reading bytes on the other end but this shouldn't
				// block since we aren't writing enough bytes to overflow the pipe buffer.
				decoded_Audio_Piped_Output_Stream.write(buffer, 0, number_Of_bytes_To_Copy);
			}
		}catch(Exception e){}


		// Send a notification to the main thread that the audio processing thread is now ready to start having data
		// read from the pipe and/or to start playing audio
		synchronized(this){
			try{
				is_Ready_To_Proceed=true;
				this.notify();
			}catch(Exception exception){}
		}


		while(!should_Exit){
			// Start audio playback if necessary.
			if(should_Play_Audio && audio_Line_To_Play_Audio_To!=null && !audio_Line_To_Play_Audio_To.isRunning())
				audio_Line_To_Play_Audio_To.start();


			try{
				// Get more audio data if all previous audio data has been written.
				if(number_Of_Bytes_Written==number_Of_Bytes_Read){
					number_Of_Bytes_Read=decoded_Audio_Input_Stream.read(buffer);
					number_Of_Bytes_Written=0;
				}


				// If number_Of_Bytes_Read is -1 that means the end of the decoded_Audio_Input_Stream was reached so
				// we should exit by setting should_Exit to true.
				if(number_Of_Bytes_Read==-1)
					should_Exit=true;


				// Write read audio data to the audio_Line_To_Play_Audio_To and decoded_Audio_Piped_Output_Stream as
				// necessary until all the read data is written.
				while(number_Of_Bytes_Written<number_Of_Bytes_Read){
					number_Of_bytes_To_Copy=number_Of_Bytes_Read-number_Of_Bytes_Written;

					// Write the audio data to audio_Line_To_Play_Audio_To if necessary.
					if(audio_Line_To_Play_Audio_To!=null)
						number_Of_bytes_To_Copy=audio_Line_To_Play_Audio_To.write(buffer, number_Of_Bytes_Written,
						  number_Of_bytes_To_Copy);

					// Write the audio data to decoded_Audio_Piped_Output_Stream.
					decoded_Audio_Piped_Output_Stream.write(buffer, number_Of_Bytes_Written, number_Of_bytes_To_Copy);

					number_Of_Bytes_Written+=number_Of_bytes_To_Copy;
				}
			}catch(Exception e){
				System.out.println("An exception occurred in the audio processing thread.");
				break;
			}
		}


		// Shutdown and cleanup.
		if(audio_Line_To_Play_Audio_To!=null){
			if(audio_Line_To_Play_Audio_To.isRunning())
				audio_Line_To_Play_Audio_To.stop();
			if(audio_Line_To_Play_Audio_To.isOpen())
				audio_Line_To_Play_Audio_To.close();
		}

		try{
			decoded_Audio_Piped_Output_Stream.close();
			decoded_Audio_Input_Stream.close();
		}catch(Exception e){}
	}
}